/*---------------------------------------------------------------
 *	PWMを使ってLEDの明るさを徐々に変化させる(μT-Kernel 3.0用)
 *
 *	Copyright (C) 2022-2024 by T3 WG of TRON Forum
 *---------------------------------------------------------------*/
#include <tk/tkernel.h>
#include <tm/tmonitor.h>

//-------- PWMのレジスタ定義 ------------------------------------
#define PWM_BASE(un)	((un) == 0 ? 0x4001c000 : (un) == 1 ? 0x40021000 : (un) == 2 ? 0x40022000 : 0x4002d000)
#define PWM(un, r)				(PWM_BASE(un) + PWM_##r)
#define PWM_TASKS_STOP			0x004
#define PWM_TASKS_SEQSTART(n)	(0x008 + (n) * 0x04)
#define PWM_LOOPSDONE			0x11C
#define PWM_ENABLE				0x500
#define PWM_MODE				0x504
#define PWM_COUNTERTOP			0x508
#define PWM_PRESCALER			0x50c
#define PWM_DECODER				0x510
#define PWM_LOOP				0x514
#define PWM_SEQ_PTR(n)			(0x520 + (n) * 0x20)
#define PWM_SEQ_CNT(n)			(0x524 + (n) * 0x20)
#define PWM_SEQ_REFRESH(n)		(0x528 + (n) * 0x20)
#define PWM_SEQ_ENDDELAY(n)		(0x52c + (n) * 0x20)
#define PWM_PSEL_OUT(n)			(0x560 + (n) * 0x04)

//-------- LED制御のためのGPIOの初期設定 ------------------------
LOCAL void led_init(void)
{
	// COL1..COL5とROW1..ROW5に接続されたGPIOピンを出力に設定
	out_w(GPIO(P0, PIN_CNF(28)), 1);	// GPIO P0.28(COL1に接続)
	out_w(GPIO(P0, PIN_CNF(11)), 1);	// GPIO P0.11(COL2に接続)
	out_w(GPIO(P0, PIN_CNF(31)), 1);	// GPIO P0.31(COL3に接続)
	out_w(GPIO(P1, PIN_CNF(05)), 1);	// GPIO P1.05(COL4に接続)
	out_w(GPIO(P0, PIN_CNF(30)), 1);	// GPIO P0.30(COL5に接続)
	out_w(GPIO(P0, PIN_CNF(21)), 1);	// GPIO P0.21(ROW1に接続)
	out_w(GPIO(P0, PIN_CNF(22)), 1);	// GPIO P0.22(ROW2に接続)
	out_w(GPIO(P0, PIN_CNF(15)), 1);	// GPIO P0.15(ROW3に接続)
	out_w(GPIO(P0, PIN_CNF(24)), 1);	// GPIO P0.24(ROW4に接続)
	out_w(GPIO(P0, PIN_CNF(19)), 1);	// GPIO P0.19(ROW5に接続)
}

//-------- GPIOの特定のピンpinの出力をval(0または1)に設定 -------
//	pinで指定された以外のピンの状態は変化しない
LOCAL void out_gpio_pin(UW port, UW pin, UW val)
{
	INT port_addr;

	if(port == 1){							// P1の場合
		if(val)
			port_addr = GPIO(P1, OUTSET);	// P1でvalが1の場合
		else
			port_addr = GPIO(P1, OUTCLR);	// P1でvalが0の場合

	} else {								// P0の場合
		if(val)
			port_addr = GPIO(P0, OUTSET);	// P0でvalが1の場合
		else
			port_addr = GPIO(P0, OUTCLR);	// P0でvalが0の場合
	}
	out_w(port_addr, (1 << pin));			// 指定のピンに1を出力
}

// カウンタ上限値とカウンタ比較値を定義するDUTYマクロ(※E)
#define	TOP				1000
#define	DUTY(percent)	(TOP - (TOP * percent / 100))

// シーケンス0のカウンタ比較値を格納する16ビット幅の配列(※G)
static _UH pwm_seq0_comp[] = {
	DUTY(0),   DUTY(10), DUTY(50), DUTY(100),
	DUTY(5),   DUTY(14), DUTY(40), DUTY(70),
	DUTY(15),  DUTY(19), DUTY(32), DUTY(50),
	DUTY(30),  DUTY(25), DUTY(25), DUTY(30),
	DUTY(50),  DUTY(32), DUTY(19), DUTY(15),
	DUTY(70),  DUTY(40), DUTY(14), DUTY(5),
	DUTY(100), DUTY(50), DUTY(10), DUTY(0)
};

// シーケンス1のカウンタ比較値を格納する16ビット幅の配列(※H)
static _UH pwm_seq1_comp[] = {
	DUTY(100), DUTY(50), DUTY(10), DUTY(0),
	DUTY(50),  DUTY(50), DUTY(10), DUTY(0),
	DUTY(10),  DUTY(10), DUTY(10), DUTY(0),
	DUTY(1),   DUTY(1),  DUTY(1), DUTY(0),
	DUTY(0),   DUTY(0),  DUTY(0), DUTY(0)
};

// unitのPWMに上限値とシーケンスの比較値を設定して動作開始(※J)
LOCAL void pwm_start_seq(INT unit, INT loopcnt, INT ref0, INT ref1)
{
	tm_printf("pwm_start_seq: unit=%d, loop=%d, seq0ref=%d, seq1ref=%d\n",
			unit, loopcnt, ref0, ref1);

	out_w(PWM(unit, ENABLE), (1 << 0));
	out_w(PWM(unit, MODE), 0);
	out_w(PWM(unit, PRESCALER), 4);		// DIV_16でクロック1MHzを指定
	out_w(PWM(unit, COUNTERTOP), TOP);	// カウンタ上限値の設定
	out_w(PWM(unit, DECODER), 2);		// LOAD=individual(single)を指定(※D)

	// シーケンス0のカウンタ比較値の配列を入れたメモリアドレス(SEQ_PTR(0))、
	//				配列のデータ数(SEQ_CNT(0))、REFRESHレジスタの設定(※K)
	out_w(PWM(unit, SEQ_PTR(0)), (UW)&pwm_seq0_comp);
	out_w(PWM(unit, SEQ_CNT(0)), (sizeof(pwm_seq0_comp) / sizeof(_UH)));
	out_w(PWM(unit, SEQ_REFRESH(0)), ref0);
	out_w(PWM(unit, SEQ_ENDDELAY(0)), 0);

	// シーケンス1のカウンタ比較値の配列を入れたメモリアドレス(SEQ_PTR(1))、
	//				配列のデータ数(SEQ_CNT(1))、REFRESHレジスタの設定(※K)
	out_w(PWM(unit, SEQ_PTR(1)), (UW)&pwm_seq1_comp);
	out_w(PWM(unit, SEQ_CNT(1)), (sizeof(pwm_seq1_comp) / sizeof(_UH)));
	out_w(PWM(unit, SEQ_REFRESH(1)), ref1);
	out_w(PWM(unit, SEQ_ENDDELAY(1)), 0);

	// LOOPレジスタに繰り返し回数を設定(※L)
	out_w(PWM(unit, LOOP), loopcnt);

	// PWMのシーケンス0の動作を開始
	out_w(PWM(unit, TASKS_SEQSTART(0)), 1);
}

const INT led_pwm = 0;				 	// LED用PWMとしてunit=0を使用

EXPORT void usermain(void)
{
	led_init();

	// LEDを点灯させるためのGPIOの設定(※F)
	out_gpio_pin(0, 28, 0);		// GPIO P0.28-COL1に0を出力
	out_gpio_pin(0, 11, 0);		// GPIO P0.11-COL2に0を出力
	out_gpio_pin(0, 31, 0);		// GPIO P0.31-COL3に0を出力
	out_gpio_pin(1,  5, 0);		// GPIO P1.05-COL4に0を出力
	out_gpio_pin(0, 30, 0);		// GPIO P0.30-COL5に0を出力

    // PWMの各チャンネル出力をGPIOピン経由でROW5..ROW2に出力(※A)
	out_w(PWM(led_pwm, PSEL_OUT(0)), ((0 << 5) | (19 << 0)));
										// チャンネル0→P0.19(ROW5に接続)
	out_w(PWM(led_pwm, PSEL_OUT(1)), ((0 << 5) | (24 << 0)));
										// チャンネル1→P0.24(ROW4に接続)
	out_w(PWM(led_pwm, PSEL_OUT(2)), ((0 << 5) | (15 << 0)));
										// チャンネル2→P0.15(ROW3に接続)
	out_w(PWM(led_pwm, PSEL_OUT(3)), ((0 << 5) | (22 << 0)));
										// チャンネル3→P0.22(ROW2に接続)

	const INT loopcnt = 5;
	INT seq0_refresh = 50;	INT seq1_refresh = 100;

	// PWMのシーケンス機能の設定と動作開始
	pwm_start_seq(led_pwm, loopcnt, seq0_refresh, seq1_refresh);

	tk_slp_tsk(TMO_FEVR);			// 永久待ち
}
